unit R_WGS84;
//**************************************************************************
//    R_WGS84 - rutiny pro pevody JTSK - WGS-84
//--------------------------------------------------------------------------
//    Datum vytvoen   : 17.6.2001
//    Sprvce           : Jakub Kerhat
//    Soust projektu  : <CommRtn>
//--------------------------------------------------------------------------
//    Copyright (c) 2001  CDSw - City Data Software spol. s r.o.
//**************************************************************************
//    $Id: R_WGS84.pas,v 1.16 2004/11/29 14:28:56 jakub Exp $

interface

{MapServer-compliant}

uses T_Coord;

function WGS84toJTSK(Latitude, Longitude: Double; var A: TMapCoord): Boolean;
  // vraci JTSK souadnice pro zadanou severn ku a vchodn dlku
  // zmenou GPS v souadnm systmu elipsoidu WGS84

function BesseltoJTSK(Latitude, Longitude: Double; var C: TMapCoord): Boolean;
  // vraci JTSK souadnice pro zadanou Severn ku a vchodn dlku
  // nen teba odetat Ferro

function Convergention(const A: TMapCoord): TMapAngle;
  // hel mezi severem na besselov elipsoidu a severem v JTSK v bode A

function WGS84toBessel(var Latitude, Longitude, Altitude: Double): Boolean; overload;
function WGS84toBessel(var Latitude, Longitude: Double) : Boolean; overload;
  // konverze z elipsoidu WGS84 do Besselova elipsoidu bez Ferra

function JTSKtoWGS84(JTSK: TMapCoord; var Latitude, Longitude: Double): Boolean;
  // zptn konverze z JTSK do WGS84 (iterace - pomalej)

type TWgs84Format = (afMinutes, afSeconds);
//                 XX XX.XXX'  XX XX' XX.X''

function WGS84ToStr(Latitude, Longitude: Double; AngleFormat: TWGS84Format = afSeconds): String; overload;
function WGS84ToStr(Angle: Double; AngleFormat: TWGS84Format = afSeconds): String; overload;
// pevede hel na textov etzec
function StrToWGS84(Str: String): TMapAngle;
// textov etzec na hel

// Latitude, Longitude jsou ve stupnch!

const Ferro = -17-40/60;
  // hel, kter se za Kovka pipotval k  vchodn dlce tak aby dlkov
  // souadnice pro Evropu, Afriku a Asii byly v kladnm kvadrantu

implementation
//******************************************************************************

uses Math, SysUtils;

function DistPoints(const P1,P2: TMapCoord): Double;
//******************************************************************************
begin
  Result := Hypot(P1.X-P2.X, P1.Y-P2.Y);
  if (Result < EPS) then Result := 0;
end;

function BesseltoJTSK;
//******************************************************************************
const a     = 6377397.15508;
      e     = 0.081696831215303;
      n     = 0.97992470462083;
      rho_0 = 12310230.12797036;
      sinUQ = 0.863499969506341;
      cosUQ = 0.504348889819882;
      sinVQ = 0.420215144586493;
      cosVQ = 0.907424504992097;
      alfa  = 1.000597498371542;
      k_2   = 1.00685001861538;
var rho, eps, V, t, B, L : Double;
    sinS, cosS, sinU, cosU, cosDV, sinDV, sinV, cosV, sinB, sinD, cosD : Double;
begin
  Result := True;
  try
    B := DegToRad(Latitude);
    L := DegToRad(Longitude);

    sinB := sin(B);
    t := (1-e*sinB)/(1+e*sinB);
    t := sqr(1+sinB)/(1-sqr(sinB)) * exp(e*ln(t));
    t := k_2 * exp(alfa*ln(t));

    sinU  := (t-1)/(t+1);
    cosU  := sqrt(1-sinU*sinU);
    V     := alfa*L;
    sinV  := sin(V);
    cosV  := cos(V);
    cosDV := cosVQ*cosV + sinVQ*sinV;
    sinDV := sinVQ*cosV - cosVQ*sinV;
    sinS  := sinUQ*sinU + cosUQ*cosU*cosDV;
    cosS  := sqrt(1-sinS*sinS);
    sinD  := sinDV*cosU/cosS;
    cosD  := sqrt(1-sinD*sinD);

    eps := n*arctan(sinD/cosD);
    rho := rho_0*exp(-n*ln((1+sinS)/cosS));

    C.Y := rho*sin(eps);
    C.X := rho*cos(eps);
  except
    Result := False;
  end;
end;


function Convergention;
//******************************************************************************
begin
  IF (A.X <> 0.0)
    then Result := DegToRad(0.000008257 * A.Y + 2.373 * A.Y/A.X)
    else Result := 0;
end;

function WGS84toBessel(var Latitude, Longitude, Altitude: Double): Boolean;
//******************************************************************************
var xl, yl, zl, x2, y2, z2: double;
  B, L, H: double;

procedure Transformace(xs, ys, zs: Double; var xn, yn, zn: Double);
//------------------------------------------------------------------------------
// transformace pravouhlych souradnic
const
  // koeficienty transformace ze systemu WGS-84 do systemu S-JTSK
  dx = -570.69; dy = -85.69; dz = -462.84; {posunuti}
  wx = 4.99821/3600*pi/180; wy = 1.58676/3600*pi/180; wz = 5.2611/3600*pi/180; {rotace}
  m  = -3.543e-6; {meritko}
begin
  xn := dx + (1+m)*(+   xs + wz*ys - wy*zs);
  yn := dy + (1+m)*(-wz*xs +    ys + wx*zs);
  zn := dz + (1+m)*(+wy*xs - wx*ys +    zs);
end;

procedure BLH_xyz(B, L, H: Double; var x, y, z: Double);
//------------------------------------------------------------------------------
// vypocet pravouhlych souradnic z geodetickych souradnic
const
// parametry elipsoidu WGS-84
  a   = 6378137.0;
  f_1 = 298.257223563;
var
  rho, e2: Double;
begin
  e2  := 1 - sqr(1-1/f_1);
  rho := a/sqrt(1-e2*sqr(sin(B)));
  x := (rho+H) * cos(B)*cos(L);
  y := (rho+H) * cos(B)*sin(L);
  z := ((1-e2)*rho+H) * sin(B);
end;

procedure xyz_BLH(x, y, z: Double; var B, L, H: Double);
//------------------------------------------------------------------------------
// vypocet geodetickych souradnic z pravouhlych souradnic
const
  a   = 6377397.15508;  // parametry Besselova elipsoidu
  f_1 = 299.152812853;
var
  a_b, e2, th, st, ct, p, t: Double;
begin
  a_b := f_1/(f_1-1);
  p   := sqrt(sqr(x) + sqr(y));
  e2  := 1-sqr(1-1/f_1);
  th  := arctan(z*a_b/p);
  st  := sin(th);
  ct  := cos(th);
  t   := (z + e2*a_b*a*IntPower(st,3))/(p - e2*a*IntPower(ct,3));

  B := arctan(t);
  H := sqrt(1+t*t) * (p-a/sqrt(1+(1-e2)*t*t));
  L := 2*arctan(y/(p+x));
end;

//------------------------------------------------------------------------------
begin
  Result := True;
  try
    B := DegToRad(Latitude);
    L := DegToRad(Longitude);
    H := Altitude;

    BLH_xyz(B,L,H,xl,yl,zl);
    transformace(xl,yl,zl,x2,y2,z2);
    xyz_BLH(x2,y2,z2,B,L,H);

    Latitude := RadToDeg(B);
    Longitude := RadToDeg(L);
    Altitude := H;
  except
    Result:=false;
  end;
end;

function WGS84toBessel(var Latitude, Longitude: Double) : Boolean;
//******************************************************************************
var Altitude: Double;
begin
  Altitude := 0;
  Result := WGS84toBessel(Latitude, Longitude, Altitude);
end;

function WGS84toJTSK;
//******************************************************************************
begin
  if (Latitude < 40) or (Latitude > 60) or (Longitude < 5) or (Longitude > 25) then
    begin
    A := NullCoord;
    Result := false;
    end
  else
    Result := WGS84toBessel(Latitude, Longitude) AND BesseltoJtsk(Latitude, Longitude, A);
end;

function JTSKtoWGS84;
//******************************************************************************
var
  Delta: Double;
  V1, V2, V3, V4: Double;
  A: TMapCoord;
  Iterace: Integer;
begin
  if (JTSK.X = 0) and (JTSK.Y = 0) then
    begin
    Latitude := 0;
    Longitude := 0;
    Result := false;
    Exit;
    end;

  Delta := 5;
  Latitude := 49;
  Longitude := 14;

  Iterace := 0;

  repeat
    if WGS84toJTSK(Latitude - Delta, Longitude - Delta, A) then
      V1 := DistPoints(A, Jtsk)
    else
      V1 := 1e32;

    if WGS84toJTSK(Latitude - Delta, Longitude + Delta, A) then
      V2 := DistPoints(A, Jtsk)
    else
      V2 := 1e32;

    if WGS84toJTSK(Latitude + Delta, Longitude - Delta, A) then
      V3 := DistPoints(A, Jtsk)
    else
      V3 := 1e32;

    if WGS84toJTSK(Latitude + Delta, Longitude + Delta, A) then
      V4 := DistPoints(A, Jtsk)
    else
      V4 := 1e32;

    if (V1 <= V2) and (V1 <= V3) and (V1 <= V4) then
      begin
      Latitude := Latitude - Delta / 2;
      Longitude := Longitude - Delta / 2;
      end;

    if (V2 <= V1) and (V2 <= V3) and (V2 <= V4) then
      begin
      Latitude := Latitude - Delta / 2;
      Longitude := Longitude + Delta / 2;
      end;

    if (V3 <= V1) and (V3 <= V2) and (V3 <= V4) then
      begin
      Latitude := Latitude + Delta / 2;
      Longitude := Longitude - Delta / 2;
      end;

    if (V4 <= V1) and (V4 <= V2) and (V4 <= V3) then
      begin
      Latitude := Latitude + Delta / 2;
      Longitude := Longitude + Delta / 2;
      end;

    Delta := Delta * 0.55;
    inc(Iterace, 4);

  until (Delta < 0.00001) or (Iterace > 1000);

  Result := Iterace < 1000;
end;


function WGS84ToStr(Latitude, Longitude: Double; AngleFormat: TWGS84Format = afSeconds): String;
//******************************************************************************
begin
  case AngleFormat of
    afMinutes: begin
      Latitude := Round(Latitude * 60 * 1000) / 60 / 1000;
      Longitude := Round(Longitude * 60 * 1000) / 60 / 1000;
      Result := Format('[N%d %0.3f'', E%d %0.3f'']',
        [Trunc(Latitude), Round(Frac(Latitude) * 60000) / 1000,
        Trunc(Longitude), Round(Frac(Longitude) * 60000) / 1000]);
      end;
    afSeconds: begin
      Latitude := Round(Latitude * 60 * 600) / 60 / 600;
      Longitude := Round(Longitude * 60 * 600) / 60 / 600;
      Result := Format('[N%d %d'' %0.1f'''', E%d %d'' %0.1f'''']',
        [Trunc(Latitude), Trunc(Frac(Latitude) * 60), Round(Frac(Latitude * 60) * 600) / 10,
        Trunc(Longitude), Trunc(Frac(Longitude) * 60), Round(Frac(Longitude * 60) * 600) / 10]);
      end;
    end;
end;

function WGS84ToStr(Angle: Double; AngleFormat: TWGS84Format = afSeconds): String;
//******************************************************************************
begin
  case AngleFormat of
    afMinutes: begin
      Angle := Round(Angle * 60 * 1000) / 60 / 1000;
      Result := Format('%d %0.3f''',
        [Trunc(Angle), Round(Frac(Angle)*60000)/1000]);
      end;
    afSeconds: begin
      Angle := Round(Angle * 60 * 600) / 60 / 600;
      Result := Format('%d %d'' %0.1f''''',
        [Trunc(Angle), Trunc(Frac(Angle)*60), Round(Frac(Angle*60)*600)/10]);
      end;
    end;
end;

function StrToWGS84(Str: String): TMapAngle;
//******************************************************************************
var S: String;
begin
  try
    if Pos('', Str) > 0 then
      begin
      S := Trim(Copy(Str, 1, Pos('', Str) - 1));
      Str := Copy(Str, Pos('', Str) + 1, MaxInt);
      Result := StrToFloat(S);
      if Pos('''', Str) > 0 then
        begin
        S := Trim(Copy(Str, 1, Pos('''', Str) - 1));
        Str := Copy(Str, Pos('''', Str) + 1, MaxInt);
        Result := Result + StrToFloat(S) / 60;
        if Pos('''''', Str) > 0 then
          begin
          S := Trim(Copy(Str, 1, Pos('''''', Str) - 1));
          Str := Copy(Str, Pos('''''', Str) + 2, MaxInt);
          Result := Result + StrToFloat(S) / 60 / 60;
          if Trim(Str) <> '' then Result := 0;
          end;
        end;
      end
    else
      Result := 0;
  except
    Result := 0;
    end;
end;

end.
